#########################
# Zope Army Knife       #
# (c) 2003,2004 NAUMEN  #
#########################
"""
Revision information:
$Id: ZAK.py,v 1.10 2003/08/27 11:20:22 kvas Exp $

ZAK class
"""
__version__ = '$Revision: 1.10 $'[11:-2]

from posixpath import split as splitpath
from re import sub
from sys import exc_info
from traceback import format_exception_only, format_tb

from Globals import DTMLFile, InitializeClass
from Acquisition import aq_inner, aq_parent
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from Persistence import PersistentMapping, Persistent
from ZODB import PersistentList

import ZAKContext
from ZAKContext import html_format
from Watch import WatchRecord

manage_addZAKForm = DTMLFile( 'dtml/addZAKForm', globals() )

def manage_addZAK( self, id, add=None, REQUEST=None ):
    """Create a new ZAK and add it to container"""
    container = self.Destination()

    if add:
        zak = ZAK( id )
        container._setObject( id, zak )

    if REQUEST is not None:
        REQUEST.RESPONSE.redirect( self.DestinationURL() + '/manage_main' )


class ZAKHistoryRecord(Persistent):
    """ZAK stored script"""
    def __init__(self, name, text):
        self.name = name
        self.text = text


class ZAK(SimpleItem):
    """Zope Army Knife

    Zope arbitrary expression evaluator.
    """

    meta_type = 'ZAK (Zope Army Knife)'
    security = ClassSecurityInfo()

    ta_cols = 80
    ta_rows = 20
    history = None
    history_order = []
    watches = {}
    
    security.declareProtected( 'Use ZAK',
                               'index_html',
                               'manage_evaluate',
                               'manage_watch',
                               'manage_showWatch',
                               'manage_configWatch',
                               'manage_showCall',
                               'listWatches',
                               'getWatch',
                               'updateSize' )

    _indexHtml = DTMLFile( 'dtml/indexHtml', globals() )
    _evaluatorForm = DTMLFile( 'dtml/evaluatorForm', globals() )

    manage_watch = DTMLFile( 'dtml/watch', globals() )
    manage_showWatch = DTMLFile( 'dtml/showWatch', globals() )
    manage_showCall = DTMLFile( 'dtml/showCall', globals() )

    manage_options = ( {'label':'Evaluator', 'action':'manage_evaluate'},
                       {'label':'Watch', 'action':'manage_watch'} ) + \
                     SimpleItem.manage_options

    def __init__( self, id='zak' ):
        self.id = id

    def manage_evaluate( self, expr='', persist=None, REQUEST=None ):
        """Evaluate any expression in any context"""
	path, method = splitpath( REQUEST.URL )
	use_zmi = method.startswith( 'manage_' )
	
        
        if expr and REQUEST.get('eval.x', None):
            lines = []
            context = aq_parent(self)

            scope, error = evaluate(expr, context, persist, REQUEST)

            if error:
                lines += [ ''.join( format_exception_only(error[0], error[1]) ),
                           ''.join( format_tb(error[2])[1:] ),
                           '' ]

            names = scope.keys()
            names.sort()

            lines.append( 'context = ' + str(context) )
            lines += [ str(name) + ' = ' + str(scope[name]) for name in names ]

            result = html_format( '\n'.join(lines) )
        else:
            result = ''

        if REQUEST.get('taller.x', None): self.ta_rows = self.ta_rows + 5
        if REQUEST.get('shorter.x', None): self.ta_rows = self.ta_rows - 5
        if REQUEST.get('wider.x', None): self.ta_cols = self.ta_cols + 5
        if REQUEST.get('narrower.x', None): self.ta_cols = self.ta_cols - 5

        if REQUEST.get('store.x', None):
            if REQUEST.get('store_name', None):
                self.storeScript(REQUEST.get('store_name', None), expr)
            else: 
                self.storeScript(REQUEST.get('stored_id', None), expr)
        if REQUEST.get('recall.x', None):
            expr = self.history.get(REQUEST.get('stored_id', None)).text
        if REQUEST.get('del.x', None):
            self.deleteScript(REQUEST.get('stored_id', None))

        template = use_zmi and self._evaluatorForm or self._indexHtml
        return template(self, REQUEST, result=result, expr=expr)

    index_html = manage_evaluate
    
    def storeScript(self, name, text):
        if not name: return
        if self.history is None:
            self.history = PersistentMapping()
            self.history_order = PersistentList.PersistentList()
        self.history[name] = ZAKHistoryRecord(name, text)
        try:
            self.history_order.remove(name)
        except ValueError: pass
        self.history_order.append(name)

    def deleteScript(self, name):
        if not name: return
        if self.history is None: return
        try:
            del self.history[name]
            self.history_order.remove(name)
        except ValueError: pass
        except KeyError: pass

    def listWatches(self):
        """Return a list of all active watches"""
        return self.watches.items()

    def addWatch(self, method_name):
        if self.watches == {}:
            self.watches = PersistentMapping()
        self.watches[method_name] = WatchRecord(method_name)

    def delWatch(self, method_name):
        self.watches[method_name].deactivate()
        del self.watches[method_name]

    def getWatch(self, method_name):
        return self.watches[method_name]

    def manage_watches(self, REQUEST=None):
        """Do watch operations"""
        if REQUEST.get('delete.x', None):
	        for id in REQUEST.get('ids', []): self.delWatch(id)
        elif REQUEST.get('activate.x', None):
	        for id in REQUEST.get('ids', []): self.getWatch(id).activate()
        elif REQUEST.get('deactivate.x', None):
	        for id in REQUEST.get('ids', []): self.getWatch(id).deactivate()
        elif REQUEST.get('method_name', None):
            self.addWatch(REQUEST.get('method_name', None))
            
        if REQUEST:
            REQUEST.response.redirect(self.absolute_url() + '/manage_watch')

    def manage_configWatch(self, watch_id=None, REQUEST=None):
        """Change watch configuration"""
        if not watch_id:
            REQUEST.response.redirect(self.absolute_url() + '/manage_watch')
        elif REQUEST.get('delete.x', None):
            self.delWatch(watch_id)
            REQUEST.response.redirect(self.absolute_url() + '/manage_watch')
        else:
            watch = self.getWatch(watch_id)
            if REQUEST.get('clear.x', None): watch.clear()
            elif REQUEST.get('save.x', None):
                watch.setPreScript(REQUEST.get('pre_script', ''))
                watch.setPostScript(REQUEST.get('post_script', ''))
            elif REQUEST.get('activate.x', None): watch.activate()
            elif REQUEST.get('deactivate.x', None): watch.deactivate()
                
            REQUEST.response.redirect(self.absolute_url() + \
                                '/manage_showWatch?watch_id=' + str(watch_id))


InitializeClass( ZAK )


def evaluate( expr, context=None, persist=None, REQUEST=None, **params ):
    expr  = sub( r'\r+\n?', '\n', expr.expandtabs() ) + '\n'
    error = None

    if persist:
        scope = ZAKContext.context
    else:
        scope = {}

    # NB overrides 'context' variable
    scope.update( vars( ZAKContext ) )
    scope.update( params )

    scope['REQUEST'] = REQUEST
    scope['context'] = context
    scope['app']     = context.getPhysicalRoot()

    try:
        exec expr in scope
    except:
        error = exc_info()
        get_transaction().abort()

    for name in dir(ZAKContext) + params.keys() + ['REQUEST','context','app']:
        try: del scope[ name ]
        except KeyError: pass

    return ( scope, error )
